#include <avr/power.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // For any tables in ROM


/* Old Crow's DDS A440 references
 
   Martin Nawrath -- Arduino DDS sine wave generator example code

   Revision: 0.1
   LAST MODIFIED 06-OCT-2015 20:19

   Generates a 440.0Hz PWM sine tone, gated by A440SW. Uses Timer 0.

   Fuses should be set to L:0xFF H:0xDD E:0xFF LB:0x3F
*/

#define  bGrab  pgm_read_byte_near // Define shorter names for these progmem calls
#define  wGrab  pgm_read_word_near //  "   "

#if !defined(USART0_RX_vect) && defined(USART0__RX_vect)
    #define USART0_RX_vect USART0__RX_vect
#endif

//
// Pin Definitions

const int a440pl = 2;  // TINY85 pin 7 (PB2), Switch polarity jumper
const int a440pw = 1;  // TINY85 pin 5 (PB1), PWM output for A440 sine tone
const int a440sw = 0;  // TINY85 pin 6 (PB0), A440 on/off switch input
// Lookup tables in program flash area
// Table of 256 sine values stored as one sine period.
const uint8_t sineTable[] PROGMEM = {
  127,130,133,136,139,143,146,149,152,155,158,161,164,167,170,173,176,178,181,184,187,190,192,195,
  198,200,203,205,208,210,212,215,217,219,221,223,225,227,229,231,233,234,236,238,239,240,242,243,
  244,245,247,248,249,249,250,251,252,252,253,253,253,254,254,254,254,254,254,254,253,253,253,252,
  252,251,250,249,249,248,247,245,244,243,242,240,239,238,236,234,233,231,229,227,225,223,221,219,
  217,215,212,210,208,205,203,200,198,195,192,190,187,184,181,178,176,173,170,167,164,161,158,155,
  152,149,146,143,139,136,133,130,127,124,121,118,115,111,108,105,102,99,96,93,90,87,84,81,78,76,73,
  70,67,64,62,59,56,54,51,49,46,44,42,39,37,35,33,31,29,27,25,23,21,20,18,16,15,14,12,11,10,9,7,6,5,
  5,4,3,2,2,1,1,1,0,0,0,0,0,0,0,1,1,1,2,2,3,4,5,5,6,7,9,10,11,12,14,15,16,18,20,21,23,25,27,29,31,
  33,35,37,39,42,44,46,49,51,54,56,59,62,64,67,70,73,76,78,81,84,87,90,93,96,99,102,105,108,111,115,
  118,121,124
};

// Variables
boolean a440Switch, a440SwLast, a440SwPol; // For reading the A440 physical switch and polarity jumper
// DDS sine variables
double dfreq;  // Set up dfreq to be adjustable w/default of 440.0Hz (assigned below)
// const double refclk = 32207.06; // Measured
const double refclk = 31376.6; // =16MHz / 510
// const double refclk = 15686.3; // =8Mhz / 510
// Variables used inside DDS interrupt service routine declared as volatile.
volatile byte indexOffset;        // Offset into sine table
volatile unsigned long phaseAcc;  // phase accumulator
volatile unsigned long tuneWord;  // dds tuning word

void disableA440() {
      TIMSK &= ~(1 << TOIE0); // Disable timer 0 interrupt (DDS sine)
      OCR0B = 0; // Mute PWM carrier
}

void enableA440() {
      TIMSK |= (1 << TOIE0);  // Enable timer 0 interrupt (DDS sine)
}
  
// Timer0 interrupt service every 1/31372.550 Hz = 32uSec
// This is the timebase reference for the DDS generator
// Freq = (TUNEWORD * REFCLK) / (2 exp 32)

// Avoid ISR(TIMER0) duplicate declaration issues with the stock Arduino delay routines
// by using int main() instead of void setup() and void loop().  This means delay(), millis()
// and so on are not available.

ISR(TIMER0_OVF_vect) {
  phaseAcc = phaseAcc + tuneWord;      // soft DDS, phase accumulator with 32 bits
  indexOffset = phaseAcc >> 24;       // use upper 8 bits of phase accumulator as waveform index
  OCR0B = bGrab(sineTable + indexOffset); // grab value from sine table and send to PWM DAC
}

// No void setup() or void loop() used to keep compiler from trying to link to built-in timer0 ISR
int main(void) {

  	noInterrupts();
        if(F_CPU == 16000000) clock_prescale_set(clock_div_1); // USe 16MHz crystal
        pinMode(a440pl, INPUT);
        pinMode(a440pw, OUTPUT);
        pinMode(a440sw, INPUT);
        
// Timer0 setup
// Set prescaler to 1, PWM mode to phase correct PWM,  8M/510 (16M/510) = 15686.3 Hz (31372.55 Hz) clock

        TCCR0A = 0;
        TCCR0B = 0;
        TCNT0 = 0;
        TCCR0B |= (1 << CS00);    // Set timer0 clock prescaler to 1
        TCCR0A |= (1 << COM0B1);  // /
        TCCR0A |= (1 << WGM00);   // Mode 1  / Phase Correct PWM
        TIMSK &= ~(1 << TOIE0);  // Disable timer 2 interrupt by default.

        dfreq=440.0;              // initial output frequency = 440.0 Hz
        tuneWord = pow(2,32) * dfreq / refclk; // calulate DDS tuning word 

        a440Switch = digitalRead(a440sw);  // Initial switch position
        a440SwLast = ~a440Switch;
        interrupts();
        
        while(1) { // void loop();
                for (int i=0;i<100;i++) {}; //Extremely cheesy debounce delay
                a440Switch = digitalRead(a440sw);
                if (a440Switch != a440SwLast) { // Set/reset only on state change
                    a440SwPol = a440Switch; // Default switch action is left=on
                    if(digitalRead(a440pl)) a440SwPol = ~a440SwPol; // Jumper open means right=on
                    a440SwPol ? disableA440(): enableA440();
                }
//                        a440Switch ? TIMSK |= (1 << TOIE0): TIMSK &= ~(1 << TOIE0);
                a440SwLast = a440Switch; // Note state for toggle
        }
}    
// END

